/*****************************************************************************
 *  FILE   : calc_Chat_itMinC.c
 *  AUTHOR : Jens Stoye
 *  DATE   : 03/1996
 *  RCS-ID : $Id: calc_Chat_itMinC.c,v 1.1 1997/07/11 11:21:41 jones Exp jones $
 *****************************************************************************/

/*****************************************************************************
 * GLOBAL FUNCTIONS
 *****************************************************************************/
void calc_Chat(tuple l, tuple u, tuple cut)
{
  extern sequences seqs;
  int iteration,p,q,r,o,d;
  double min,sum;
#ifdef WRITE_STATISTICS
  int rlen,x;
#endif

  /* start in the middle of each sequence */
  for(p=0; p<seqs.num; ++p)
    cut[p] = (u[p]-l[p])/2; /* 0<=cut<u-l, but cut=u-l is relevant, too */

  /* find Nullstellen in columns col[p][0] */
  for(p=1; p<seqs.num; ++p) {
    score[p][0] = calc_column(p,0,l[p],l[0],u[p],u[0],l[0]+cut[0],col[p][0]);
    for(o=0; col[p][0][cut[p]+o]!=0 && col[p][0][cut[p]-o]!=0; ++o)
      /* nothing */ ;
    tmp_cut[p] = cut[p] += (col[p][0][cut[p]+o] == 0) ? o : -o;
  }

  /* calculate remaining columns, pairwise scores, and Chat */
  Chat = 0.0;
  for(p=2; p<seqs.num; ++p) /* entries q==0 are zero, anyway */
    for(q=1; q<p; ++q) {
      score[p][q] = calc_column(p,q,l[p],l[q],u[p],u[q],l[q]+cut[q],col[p][q]);
      Chat += seqs.wf[p][q]*col[p][q][cut[p]];
    }
  
  /* calculate relevant regions: 0<=lo<=uo<=u-l */
  if(Chat > 0.0) /* avoid running over the minimum */
    for(p=1; p<seqs.num; ++p) {
      for(lo[p]=0; seqs.wf[p][0]*col[p][0][lo[p]]>=Chat; ++lo[p])
        /* nothing */ ;
      for(uo[p]=u[p]-l[p]; seqs.wf[p][0]*col[p][0][uo[p]]>=Chat; --uo[p])
        /* nothing */ ;
    }
  else /* relevant regions consist of one cell, only */
    for(p=1; p<seqs.num; ++p)
      lo[p] = uo[p] = cut[p];

  /* save original offsets */
  for(p=1; p<seqs.num; ++p) {
    olo[p] = lo[p];
    olen[p] = uo[p]-lo[p]+1;
  }

  /* allocate and calculate matrices, even if Chat==0.0 (no matter) */
  for(p=2; p<seqs.num; ++p)
    for(q=1; q<p; ++q) {
      ALLOC(m[p][q],int*,olen[p]);
      for(r=0; r<olen[p]; ++r)
        ALLOC(m[p][q][r],int,olen[q]);
      calc_matrix(p,q,l[p],l[q],u[p],u[q],lo[p],lo[q],uo[p],uo[q],
                  m[p][q],score[p][q]);
    }

  /* iterative search for a better Chat */
  iteration = 0;
  if(Chat > 0.0)
    do {
      ++iteration;
      Cmin = Chat;
      for(p=1; p<seqs.num; ++p)
        for(min=DBL_MAX,d=lo[p]; d<=uo[p] && min>0.0; ++d) {
          sum = seqs.wf[p][0]*col[p][0][d];
          for(q=1; q<p; ++q)
            sum += seqs.wf[p][q]*m[p][q][d-olo[p]][tmp_cut[q]-olo[q]];
          for(++q; q<seqs.num; ++q)
            sum += seqs.wf[q][p]*m[q][p][cut[q]-olo[q]][d-olo[p]];
          if(sum<min
             || (sum==min && ABS(olen[p]/2-d) < ABS(olen[p]/2-cut[p]))) {
            min = sum;
            tmp_cut[p] = d;
          }
        } /* for d */

      /* calculate new Chat */
      Chat = 0.0;
      for(p=1; p<seqs.num; ++p) { /* entries q==0 are not necessarily zero */
        Chat += seqs.wf[p][0]*col[p][0][tmp_cut[p]];
        for(q=1; q<p; ++q)
          Chat += seqs.wf[p][q]*m[p][q][tmp_cut[p]-olo[p]][tmp_cut[q]-olo[q]];
      }

      /* copy new tmp_cut into cut (are always better or equal) */
      for(p=1; p<seqs.num; ++p)
        cut[p] = tmp_cut[p];
    } while(Chat < Cmin);

  /* make relevant regions smaller */
  if(Chat > 0.0) /* avoid running over the minimum */
    for(p=1; p<seqs.num; ++p) {
      for( ; seqs.wf[p][0]*col[p][0][lo[p]]>=Chat; ++lo[p])
        /* nothing */ ;
      for( ; seqs.wf[p][0]*col[p][0][uo[p]]>=Chat; --uo[p])
        /* nothing */ ;
    }
  else /* relevant regions consist of one cell, only */
    for(p=1; p<seqs.num; ++p)
      lo[p] = uo[p] = cut[p];

#ifdef WRITE_STATISTICS
  if(l[0]==0 && u[0]==seqs.len[0]) {
    for(rlen=0,x=1; x<seqs.num; ++x)
      rlen = MAX(rlen,uo[x]-lo[x]+1);
    rlen = Chat==0.0 ? 1 : rlen;
    fprintf(stderr," %6d Chat",(int)Chat);
    fprintf(stderr," %2d I",iteration);
    fprintf(stderr," %4d r",rlen);
    fprintf(stderr," %6.4f Rquotient",(float)rlen/(u[0]-l[0]+1));
  }
#endif /* WRITE_STATISTICS */

} /* calc_Chat() */

/****** EOF (calc_Chat_itMinC.c) *********************************************/
