/*****************************************************************************
 *  FILE   : calc_Cmin_tboundsR.c
 *  AUTHOR : Jens Stoye
 *  DATE   : 08/1996
 *  RCS-ID : $Id: calc_Cmin_tboundsR.c,v 1.1 1997/07/11 11:24:46 jones Exp jones $
 *****************************************************************************/

/*****************************************************************************
 * CONSTANTS
 *****************************************************************************/
#define RECURSION_STEP 4 /* difference between tuplesize and next_tuplesize */
#define STOP_TUPLESIZE 7 /* smallest kappa for recursive call */

/*****************************************************************************
 * MACROS
 *****************************************************************************/
#define SWAP_COLUMNS(c,c1,c2) swap_c=c[c1][0]; c[c1][0]=c[c2][0];\
                                               c[c2][0]=swap_c;
#define SWAP_MATRICES(m,m1,m2) for(swap_p=1; swap_p<seqs.num; ++swap_p) {\
                                 swap_m = m[m1][swap_p];\
                                 m[m1][swap_p] = m[m2][swap_p];\
                                 m[m2][swap_p] = swap_m;\
                               }\
                               for(swap_p=1; swap_p<seqs.num; ++swap_p) {\
                                 swap_m = m[swap_p][m1];\
                                 m[swap_p][m1] = m[swap_p][m2];\
                                 m[swap_p][m2] = swap_m;\
                               }
#define SWAP_BOUNDS(b,b1,b2) for(swap_p=1; swap_p<seqs.num; ++swap_p) {\
                               swap_b = b[b1][swap_p];\
                               b[b1][swap_p] = b[b2][swap_p];\
                               b[b2][swap_p] = swap_b;\
                             }\
                             for(swap_p=1; swap_p<seqs.num; ++swap_p) {\
                               swap_b = b[swap_p][b1];\
                               b[swap_p][b1] = b[swap_p][b2];\
                               b[swap_p][b2] = swap_b;\
                             }

/*****************************************************************************
 * LOCAL GLOBAL VARIABLES
 *****************************************************************************/
column swap_c;
matrix swap_m;
int swap_p,*swap_b;

double act_indep,*c_indep;
int *local_lo,*local_uo, *largest,*idc,*idc2;
tuple hat_cut,local_cut;

int *lcol,*ucol, /* monotony bounds in columns with sequence s_0 */
    ***ls,***us; /* monotony bounds in matrices */

/*****************************************************************************
 * LOCAL FUNCTIONS
 *****************************************************************************/
void inc_local_tmp_cut(void)
{
  if(down[i])
    if(lower_tmp_cut[i]>local_lo[i])
      tmp_cut[i] = --lower_tmp_cut[i];
    else {
      tmp_cut[i] = ++upper_tmp_cut[i];
      lower_tmp_cut[i] = local_lo[i]-1;
    }
  else
    if(upper_tmp_cut[i]<local_uo[i])
      tmp_cut[i] = ++upper_tmp_cut[i];
    else {
      tmp_cut[i] = --lower_tmp_cut[i];
      upper_tmp_cut[i] = local_uo[i]+1;
    }
  down[i] = !down[i];

  for(act_cost-=c[i];
      i>0 && lower_tmp_cut[i]<local_lo[i] && upper_tmp_cut[i]>local_uo[i];
      act_cost-=c[i]) { /* might result in truncation errors! */

    --i; /* for i==0, the following may be undefined */

    if(down[i])
      if(lower_tmp_cut[i]>local_lo[i])
        tmp_cut[i] = --lower_tmp_cut[i];
      else {
        tmp_cut[i] = ++upper_tmp_cut[i];
        lower_tmp_cut[i] = local_lo[i]-1;
      }
    else
      if(upper_tmp_cut[i]<local_uo[i])
        tmp_cut[i] = ++upper_tmp_cut[i];
      else {
        tmp_cut[i] = --lower_tmp_cut[i];
        upper_tmp_cut[i] = local_uo[i]+1;
      }
    down[i] = !down[i];

  } /* for act_cost */
} /* inc_local_tmp_cut() */

/*---------------------------------------------------------------------------*/
void inc_local_indep_tmp_cut(void)
{
  if(down[i])
    if(lower_tmp_cut[i]>local_lo[i])
      tmp_cut[i] = --lower_tmp_cut[i];
    else {
      tmp_cut[i] = ++upper_tmp_cut[i];
      lower_tmp_cut[i] = local_lo[i]-1;
    }
  else
    if(upper_tmp_cut[i]<local_uo[i])
      tmp_cut[i] = ++upper_tmp_cut[i];
    else {
      tmp_cut[i] = --lower_tmp_cut[i];
      upper_tmp_cut[i] = local_uo[i]+1;
    }
  down[i] = !down[i];

  for(act_cost-=c[i], act_indep-=c_indep[i];
      i>0 && lower_tmp_cut[i]<local_lo[i] && upper_tmp_cut[i]>local_uo[i];
      act_cost-=c[i], act_indep-=c_indep[i]) { /* truncation errors! */

    --i; /* for i==0, the following may be undefined */

    if(down[i])
      if(lower_tmp_cut[i]>local_lo[i])
        tmp_cut[i] = --lower_tmp_cut[i];
      else {
        tmp_cut[i] = ++upper_tmp_cut[i];
        lower_tmp_cut[i] = local_lo[i]-1;
      }
    else
      if(upper_tmp_cut[i]<local_uo[i])
        tmp_cut[i] = ++upper_tmp_cut[i];
      else {
        tmp_cut[i] = --lower_tmp_cut[i];
        upper_tmp_cut[i] = local_uo[i]+1;
      }
    down[i] = !down[i];

  } /* for act_cost */
} /* inc_local_indep_tmp_cut() */

/*---------------------------------------------------------------------------*/
double min_add_cost(int tuplesize)
{
  extern sequences seqs;
  int first_dep, next_tuplesize,
      n,o, p,q,r;
  double local_Cmin,local_Chat, next_Cmin;

  next_tuplesize = tuplesize-RECURSION_STEP;
  next_Cmin = tuplesize>=STOP_TUPLESIZE ? min_add_cost(next_tuplesize) : 0.0;

  /* sort largest (tuplesize) sequences into idc */
  for(n=1; n<tuplesize; ++n) {
    for(o=n; o>1 && largest[n]<idc[o-1]; --o) /* (bubble)sort */
      idc[o] = idc[o-1];
    idc[o] = largest[n];
  }
  /* swap largest (tuplesize) sequences to the beginning */
  for(p=1; p<tuplesize; ++p) {
    SWAP_SEQS(seqs,p,idc[p]);
    SWAP_TUPLE(olo,p,idc[p]);
    SWAP_TUPLE(lo,p,idc[p]);
    SWAP_TUPLE(uo,p,idc[p]);
    SWAP_TUPLE(hat_cut,p,idc[p]);
    SWAP_TUPLE(local_lo,p,idc[p]);
    SWAP_TUPLE(local_uo,p,idc[p]);
    SWAP_TUPLE(local_cut,p,idc[p]);
    SWAP_COLUMNS(col,p,idc[p]);
    SWAP_MATRICES(m,p,idc[p]);
    SWAP_TUPLE(lcol,p,idc[p]);
    SWAP_TUPLE(ucol,p,idc[p]);
    SWAP_BOUNDS(ls,p,idc[p]);
    SWAP_BOUNDS(us,p,idc[p]);
  }

  /* initialize local_cut and calculate local_Chat as partial sum of Chat */
  local_Chat = 0.0;
  local_cut[0] = hat_cut[0];
  for(p=1; p<tuplesize; ++p) { /* entries q==0 are not necessarily zero */
    local_cut[p] = hat_cut[p];
    local_Chat += seqs.wf[p][0]*col[p][0][local_cut[p]];
    for(q=1; q<p; ++q)
      local_Chat += seqs.wf[p][q]*m[p][q][local_cut[p]-olo[p]]
                                         [local_cut[q]-olo[q]];
  }

  local_Cmin = local_Chat;
  if(local_Cmin > next_Cmin) { /* otherwise local_Cmin and local_cut are
                                * already C-optimal */
    if(tuplesize > STOP_TUPLESIZE) {
      /* sort largest (next_tuplesize) sequences into idc2 */
      for(n=1; n<next_tuplesize; ++n) {
        for(o=n; o>1 && largest[n]<idc2[o-1]; --o) /* (bubble)sort */
          idc2[o] = idc2[o-1];
        idc2[o] = largest[n];
      }
      /* swap these sequences to the local end */
      first_dep = tuplesize-next_tuplesize+1;
      for(p=q=tuplesize-1,r=next_tuplesize-1; r>0; --p)
        if(idc[p] == idc2[r]) {
          SWAP_SEQS(seqs,p,q);
          SWAP_TUPLE(olo,p,q);
          SWAP_TUPLE(lo,p,q);
          SWAP_TUPLE(uo,p,q);
          SWAP_TUPLE(hat_cut,p,q);
          SWAP_TUPLE(local_lo,p,q);
          SWAP_TUPLE(local_uo,p,q);
          SWAP_TUPLE(local_cut,p,q);
          SWAP_COLUMNS(col,p,q);
          SWAP_MATRICES(m,p,q);
          SWAP_TUPLE(lcol,p,q);
          SWAP_TUPLE(ucol,p,q);
          SWAP_BOUNDS(ls,p,q);
          SWAP_BOUNDS(us,p,q);
          --q;
          --r;
        }

      /* initialize local offsets */
      for(p=1; p<first_dep; ++p) {
        for(local_lo[p]=lo[p]; seqs.wf[p][0]*col[p][0][local_lo[p]]+next_Cmin
            >=local_Cmin; ++local_lo[p]) /* nothing */ ;
        for(local_uo[p]=uo[p]; seqs.wf[p][0]*col[p][0][local_uo[p]]+next_Cmin
            >=local_Cmin; --local_uo[p]) /* nothing */ ;
      }
      for( ; p<tuplesize; ++p) {
        for(local_lo[p]=lo[p]; seqs.wf[p][0]*col[p][0][local_lo[p]]
            >=local_Cmin; ++local_lo[p]) /* nothing */ ;
        for(local_uo[p]=uo[p]; seqs.wf[p][0]*col[p][0][local_uo[p]]
            >=local_Cmin; --local_uo[p]) /* nothing */ ;
      }

      /* calculate minimal additional cost for (tuplesize) tuple */
      act_cost = 0.0;
      tmp_cut[1] = lower_tmp_cut[1] = upper_tmp_cut[1]
                 = (local_lo[1]+local_uo[1])/2;
      down[1] = FALSE;

      for(i=1; i>0 && local_Cmin>next_Cmin; ) {
        c_indep[i] = seqs.wf[i][0]*col[i][0][tmp_cut[i]];
        for(p=1; p<i; ++p)
          c_indep[i] += seqs.wf[i][p]*m[i][p][tmp_cut[i]-olo[i]]
                                             [tmp_cut[p]-olo[p]];
        act_indep += c_indep[i];

        if(act_indep+next_Cmin < local_Cmin) {
          ++i;
          tmp_cut[i] = lower_tmp_cut[i] = upper_tmp_cut[i]
                     = (local_lo[i]+local_uo[i])/2;
          down[i] = FALSE;

          if(i >= first_dep) {
            act_cost = act_indep;
            for( ; i>=first_dep && local_Cmin>next_Cmin; ) {
              c_indep[i] = 0.0;
              for(p=1; p<first_dep; ++p)
                c_indep[i] += seqs.wf[i][p]*m[i][p][tmp_cut[i]-olo[i]]
                                                   [tmp_cut[p]-olo[p]];
              c[i] = seqs.wf[i][0]*col[i][0][tmp_cut[i]]+c_indep[i];
              for( ; p<i; ++p)
                c[i] += seqs.wf[i][p]*m[i][p][tmp_cut[i]-olo[i]]
                                             [tmp_cut[p]-olo[p]];
              act_indep += c_indep[i];
              act_cost += c[i];

              if(act_indep+next_Cmin<local_Cmin && act_cost<local_Cmin) {
                ++i;
                if(i == tuplesize) { /* found better cut */
                  local_Cmin = act_cost;
                  /* save better cut */
                  for(p=1; p<tuplesize; ++p) /* cut[0] is fixed */
                    local_cut[p] = tmp_cut[p];
                  /* make local relevant regions smaller */
                  if(local_Cmin > next_Cmin) { /* avoid running over min. */
                    for(p=1; p<first_dep; ++p) {
                      for( ; seqs.wf[p][0]*col[p][0][local_lo[p]]+next_Cmin
                          >=local_Cmin; ++local_lo[p]) /* nothing */ ;
                      for( ; seqs.wf[p][0]*col[p][0][local_uo[p]]+next_Cmin
                          >=local_Cmin; --local_uo[p]) /* nothing */ ;
                    }
                    for( ; p<tuplesize; ++p) {
                      for( ; seqs.wf[p][0]*col[p][0][local_lo[p]]>=local_Cmin;
                          ++local_lo[p]) /* nothing */ ;
                      for( ; seqs.wf[p][0]*col[p][0][local_uo[p]]>=local_Cmin;
                          --local_uo[p]) /* nothing */ ;
                    }
                    --i;
                    inc_local_indep_tmp_cut();
                  } /* if local_Cmin */
                } /* if i */
                else {
                  tmp_cut[i] = lower_tmp_cut[i] = upper_tmp_cut[i]
                             = (local_lo[i]+local_uo[i])/2;
                  down[i] = FALSE;
                } /* else */
              } /* if act_indep */
              else {
                if(act_cost-c[i]+seqs.wf[i][0]*col[i][0][tmp_cut[i]]
                   >=local_Cmin) {
                  if(tmp_cut[i]<=lcol[i])
                    lower_tmp_cut[i] = local_lo[i]-1;
                  if(tmp_cut[i]>=ucol[i])
                    upper_tmp_cut[i] = local_uo[i]+1;
                }
                for(p=1; p<first_dep; ++p)
                  if(act_indep-c_indep[i]
                     +seqs.wf[i][p]*m[i][p][tmp_cut[i]-olo[i]]
                                           [tmp_cut[p]-olo[p]]
                     +next_Cmin>=local_Cmin
                     || act_cost-c[i]+seqs.wf[i][p]*m[i][p][tmp_cut[i]-olo[i]]
                                                           [tmp_cut[p]-olo[p]]
                        >=local_Cmin) {
                    if(tmp_cut[i]-olo[i]<=ls[i][p][tmp_cut[p]-olo[p]])
                      lower_tmp_cut[i] = local_lo[i]-1;
                    if(tmp_cut[i]-olo[i]>=us[i][p][tmp_cut[p]-olo[p]])
                      upper_tmp_cut[i] = local_uo[i]+1;
                  }
                for( ; p<i; ++p)
                  if(act_cost-c[i]+seqs.wf[i][p]*m[i][p][tmp_cut[i]-olo[i]]
                                                        [tmp_cut[p]-olo[p]]
                     >=local_Cmin) {
                    if(tmp_cut[i]-olo[i]<=ls[i][p][tmp_cut[p]-olo[p]])
                      lower_tmp_cut[i] = local_lo[i]-1;
                    if(tmp_cut[i]-olo[i]>=us[i][p][tmp_cut[p]-olo[p]])
                      upper_tmp_cut[i] = local_uo[i]+1;
                  }
                inc_local_indep_tmp_cut();
              } /* else */
            } /* for (i) */
          } /* if i */
        } /* if act_indep */
        else {
          if(act_indep-c_indep[i]+seqs.wf[i][0]*col[i][0][tmp_cut[i]]+next_Cmin
             >=local_Cmin) {
            if(tmp_cut[i]<=lcol[i])
              lower_tmp_cut[i] = local_lo[i]-1;
            if(tmp_cut[i]>=ucol[i])
              upper_tmp_cut[i] = local_uo[i]+1;
          }
          for(p=1; p<i; ++p)
            if(act_indep-c_indep[i]
               +seqs.wf[i][p]*m[i][p][tmp_cut[i]-olo[i]][tmp_cut[p]-olo[p]]
               +next_Cmin>=local_Cmin) {
              if(tmp_cut[i]-olo[i]<=ls[i][p][tmp_cut[p]-olo[p]])
                lower_tmp_cut[i] = local_lo[i]-1;
              if(tmp_cut[i]-olo[i]>=us[i][p][tmp_cut[p]-olo[p]])
                upper_tmp_cut[i] = local_uo[i]+1;
            }
          inc_local_indep_tmp_cut(); /* decr. act_cost here not necessary */
        } /* else */
      } /* for i */

      /* swap sequences, columns, and matrices back to intermediate order */
      for(p=r=1,q=first_dep; r<next_tuplesize; ++p)
        if(idc[p] == idc2[r]) {
          SWAP_SEQS(seqs,p,q);
          SWAP_TUPLE(olo,p,q);
          SWAP_TUPLE(lo,p,q);
          SWAP_TUPLE(uo,p,q);
          SWAP_TUPLE(hat_cut,p,q);
          SWAP_TUPLE(local_lo,p,q);
          SWAP_TUPLE(local_uo,p,q);
          SWAP_TUPLE(local_cut,p,q);
          SWAP_COLUMNS(col,p,q);
          SWAP_MATRICES(m,p,q);
          SWAP_TUPLE(lcol,p,q);
          SWAP_TUPLE(ucol,p,q);
          SWAP_BOUNDS(ls,p,q);
          SWAP_BOUNDS(us,p,q);
          ++q;
          ++r;
        }
    } /* if next_tuplesize */

    else { /* exhaustive search for C-optimal local_cut */

      /* initialize local offsets */
      for(p=1; p<tuplesize; ++p) {
        for(local_lo[p]=lo[p]; seqs.wf[p][0]*col[p][0][local_lo[p]]
            >=local_Cmin; ++local_lo[p]) /* nothing */ ;
        for(local_uo[p]=uo[p]; seqs.wf[p][0]*col[p][0][local_uo[p]]
            >=local_Cmin; --local_uo[p]) /* nothing */ ;
      }

      /* search for C-optimal cut */
      act_cost = 0.0;
      tmp_cut[1] = lower_tmp_cut[1] = upper_tmp_cut[1]
                 = (local_lo[1]+local_uo[1])/2;
      down[1] = FALSE;

      for(i=1; i>0 && local_Cmin>0.0; ) {
        c[i] = seqs.wf[i][0]*col[i][0][tmp_cut[i]];
        for(p=1; p<i; ++p)
          c[i] += seqs.wf[i][p]*m[i][p][tmp_cut[i]-olo[i]][tmp_cut[p]-olo[p]];
        act_cost += c[i];

        if(act_cost < local_Cmin) {
          ++i;
          if(i == tuplesize) { /* found better cut */
            local_Cmin = act_cost;
            /* save better cut */
            for(p=1; p<tuplesize; ++p) /* cut[0] is fixed */
              local_cut[p] = tmp_cut[p];
            /* make relevant regions smaller */
            if(local_Cmin > 0.0) { /* avoid running over the minimum */
              for(p=1; p<tuplesize; ++p) {
                for( ; seqs.wf[p][0]*col[p][0][local_lo[p]]>=local_Cmin;
                    ++local_lo[p]) /* nothing */ ;
                for( ; seqs.wf[p][0]*col[p][0][local_uo[p]]>=local_Cmin;
                    --local_uo[p]) /* nothing */ ;
              } /* for p */
              --i;
              inc_local_tmp_cut();
            } /* if local_Cmin */
          } /* if i */
          else {
            tmp_cut[i] = lower_tmp_cut[i] = upper_tmp_cut[i]
                       = (local_lo[i]+local_uo[i])/2;
            down[i] = FALSE;
          } /* else */
        } /* if act_cost */
        else {
          if(act_cost-c[i]+seqs.wf[i][0]*col[i][0][tmp_cut[i]]>=local_Cmin) {
            if(tmp_cut[i]<=lcol[i])
              lower_tmp_cut[i] = local_lo[i]-1;
            if(tmp_cut[i]>=ucol[i])
              upper_tmp_cut[i] = local_uo[i]+1;
          }
          for(p=1; p<i; ++p)
            if(act_cost-c[i]+seqs.wf[i][p]*m[i][p][tmp_cut[i]-olo[i]]
                                                  [tmp_cut[p]-olo[p]]
               >=local_Cmin) {
              if(tmp_cut[i]-olo[i]<=ls[i][p][tmp_cut[p]-olo[p]])
                lower_tmp_cut[i] = local_lo[i]-1;
              if(tmp_cut[i]-olo[i]>=us[i][p][tmp_cut[p]-olo[p]])
                upper_tmp_cut[i] = local_uo[i]+1;
            }
          inc_local_tmp_cut();
        } /* else */
      } /* for i */
    } /* else */

  } /* if local_Cmin */

#ifdef WRITE_STATISTICS
  else /* relevant regions consist of one cell, only */
    for(p=1; p<tuplesize; ++p)
      local_lo[p] = local_uo[p] = local_cut[p];
#endif

  /* swap sequences, columns, and matrices back to original order */
  for(p=tuplesize-1; p>=1; --p) {
    SWAP_SEQS(seqs,p,idc[p]);
    SWAP_TUPLE(olo,p,idc[p]);
    SWAP_TUPLE(lo,p,idc[p]);
    SWAP_TUPLE(uo,p,idc[p]);
    SWAP_TUPLE(hat_cut,p,idc[p]);
    SWAP_TUPLE(local_lo,p,idc[p]);
    SWAP_TUPLE(local_uo,p,idc[p]);
    SWAP_TUPLE(local_cut,p,idc[p]);
    SWAP_COLUMNS(col,p,idc[p]);
    SWAP_MATRICES(m,p,idc[p]);
    SWAP_TUPLE(lcol,p,idc[p]);
    SWAP_TUPLE(ucol,p,idc[p]);
    SWAP_BOUNDS(ls,p,idc[p]);
    SWAP_BOUNDS(us,p,idc[p]);
  }

  return(local_Cmin);
} /* min_add_cost() */

/*****************************************************************************
 * GLOBAL FUNCTIONS
 *****************************************************************************/
void calc_Cmin(tuple l, tuple u, tuple cut) 
{
  extern sequences seqs;
  int p,q, r,s, mem_p,mem_q, n, j;
  double max,this;
  BOOL *occurs,**occurs2;
#ifdef WRITE_STATISTICS
  int rlen, x;
  double maxC;
#endif

  /* transpose all matrices */
  for(p=2; p<seqs.num; ++p)
    for(q=1; q<p; ++q) {
      ALLOC(m[q][p],int*,olen[q]);
      for(r=0; r<olen[q]; ++r) {
        ALLOC(m[q][p][r],int,olen[p]);
        for(s=0; s<olen[p]; ++s)
          m[q][p][r][s] = m[p][q][s][r];
      }
    }

  /* compute monotony bounds of columns and ALL matrices */
  ALLOC(lcol,int,seqs.num); /* columns */
  ALLOC(ucol,int,seqs.num);
  for(p=1; p<seqs.num; ++p) {
    for(lcol[p]=lo[p];
        lcol[p]<uo[p] && col[p][0][lcol[p]]>=col[p][0][lcol[p]+1]; ++lcol[p]);
      /* nothing */ ;
    for(ucol[p]=uo[p];
        ucol[p]>lo[p] && col[p][0][ucol[p]]>=col[p][0][ucol[p]-1]; --ucol[p]);
      /* nothing */ ;
  }
  ALLOC(ls,int**,seqs.num); /* matrices */
  ALLOC(us,int**,seqs.num);
  for(p=1; p<seqs.num; ++p) {
    ALLOC(ls[p],int*,seqs.num);
    ALLOC(us[p],int*,seqs.num);
  }
  for(p=1; p<seqs.num; ++p)
    for(q=1; q<seqs.num; ++q)
      if(p != q) {
        ALLOC(ls[p][q],int,olen[q]);
        ALLOC(us[p][q],int,olen[q]);
        for(j=lo[q]-olo[q]; j<=uo[q]-olo[q]; ++j) {
          for(ls[p][q][j]=lo[p]-olo[p];
              ls[p][q][j]<uo[p]-olo[p] &&
                m[p][q][ls[p][q][j]][j]>=m[p][q][ls[p][q][j]+1][j];
              ++ls[p][q][j])
            /* nothing */ ;
          for(us[p][q][j]=uo[p]-olo[p];
              us[p][q][j]>lo[p]-olo[p] &&
                m[p][q][us[p][q][j]][j]>=m[p][q][us[p][q][j]-1][j];
              --us[p][q][j])
            /* nothing */ ;
        }
      }

  /* sort sequences according to their contribution to Chat */
  ALLOC(largest,int,seqs.num);
  ALLOC(occurs,BOOL,seqs.num);
  ALLOC(occurs2,BOOL*,seqs.num);
  for(p=0; p<seqs.num; ++p) {
    occurs[p] = FALSE;
    ALLOC(occurs2[p],BOOL,seqs.num);
    for(q=0; q<p; ++q)
      occurs2[p][q] = FALSE;
  }
  occurs[0] = TRUE; /* sequence 0 has to be taken in any case */
  for(n=1; n<seqs.num; ) {
    max = -DBL_MAX;
    for(p=1; p<seqs.num; ++p)
      for(q=0; q<p; ++q)
        if(!occurs2[p][q]) {
          this = seqs.wf[p][q]*(q==0 ? col[p][0][cut[p]]
                                     : m[p][q][cut[p]-olo[p]][cut[q]-olo[q]]);
          if(this > max) {
            max = this;
            mem_p = p;
            mem_q = q;
          }
        }
    occurs2[mem_p][mem_q] = TRUE;
    if(!occurs[mem_p]) {
      largest[n++] = mem_p;
      occurs[mem_p] = TRUE;
    }
    if(!occurs[mem_q]) {
      largest[n++] = mem_q;
      occurs[mem_q] = TRUE;
    }
  }
  for(p=0; p<seqs.num; ++p)
    FREE(occurs2[p]);
  FREE(occurs2);
  FREE(occurs);

  /* make slicing positions for Chat global, allocate memory */
  hat_cut = cut;
  ALLOC(local_cut,int,seqs.num);
  ALLOC(local_lo,int,seqs.num);
  ALLOC(local_uo,int,seqs.num);
  ALLOC(c_indep,double,seqs.num);
  ALLOC(idc,int,seqs.num);
  ALLOC(idc2,int,seqs.num);

  /* calculate C-optimal cut */
  Cmin = min_add_cost(seqs.num);

  /* copy local_cut to cut as return value */
  for(p=1; p<seqs.num; ++p) /* cut[0] is fixed */
    cut[p] = local_cut[p];
 
#ifdef WRITE_STATISTICS
  if(l[0]==0 && u[0]==seqs.len[0]) {
    for(maxC=-DBL_MAX,rlen=0,x=1; x<seqs.num; ++x) {
      maxC = MAX(maxC,seqs.wf[x][0]*col[x][0][cut[x]]);
      rlen = MAX(rlen,local_uo[x]-local_lo[x]+1);
    }
    rlen = Cmin==0.0 ? 1 : rlen;
    fprintf(stderr," %6d Cmin",(int)Cmin);
    fprintf(stderr," %5.2f Cquotient",Chat==0.0?1.0:Cmin/Chat);
    fprintf(stderr," %4.2f Cfactor",Chat==0.0?0.0:maxC/Chat);
    fprintf(stderr," ??? Cadd%d.%dR",RECURSION_STEP,STOP_TUPLESIZE);
    fprintf(stderr," %4d rOp",rlen);
    fprintf(stderr," %6.4f RquotientOp",(float)rlen/(u[0]-l[0]+1));
  }
#endif /* WRITE_STATISTICS */

  /* free memory */
  FREE(lcol);
  FREE(ucol);
  for(p=1; p<seqs.num; ++p) {
    for(q=1; q<seqs.num; ++q)
      if(p != q) {
        FREE(ls[p][q]);
        FREE(us[p][q]);
      }
    FREE(ls[p]);
    FREE(us[p]);
  }
  FREE(ls);
  FREE(us);

  for(p=2; p<seqs.num; ++p)
    for(q=1; q<p; ++q) {
      for(r=0; r<olen[q]; ++r)
        FREE(m[q][p][r]);
      FREE(m[q][p]);
    }
  FREE(largest);
  FREE(local_cut);
  FREE(local_lo);
  FREE(local_uo);
  FREE(c_indep);
  FREE(idc);
  FREE(idc2);

} /* calc_Cmin() */

/****** EOF (calc_Cmin_tboundsR.c) *******************************************/
