/*****************************************************************************
 *  FILE   : dypr.c
 *  AUTHOR : Jens Stoye
 *  DATE   : 05/1997
 *  RCS-ID : $Id: dypr.c,v 1.1 1997/07/11 11:15:47 jones Exp jones $
 *****************************************************************************/

/*****************************************************************************
 * INCLUDES
 *****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <limits.h> /* INT_MAX */

#include "global.h" /* data types,MIN(),MIN3(),ALLOC(),FREE() */
#include "dypr.h" /* function prototypes */

/*****************************************************************************
 * GLOBAL FUNCTION DEFINITIONS
 *****************************************************************************/
int calc_column(int s1,int s2, int l1,int l2, int u1,int u2, int c, column f)
{
  extern sequences seqs;
  extern scores scrs;

  sequence t1,t2;
  int len1,len2, i,j, mem,D,score,
      lendgapcost1,rendgapcost1,lendgapcost2,rendgapcost2;
  column Vf,Hf, b,Vb,Hb;

  t1 = seqs.s[s1]+l1;
  t2 = seqs.s[s2]+l2;
  len1 = u1-l1;
  len2 = u2-l2;

  lendgapcost1 = l1==0 ? scrs.endgapcost : scrs.gapcost;
  rendgapcost1 = u1==seqs.len[s1] ? scrs.endgapcost : scrs.gapcost;
  lendgapcost2 = l2==0 ? scrs.endgapcost : scrs.gapcost;
  rendgapcost2 = u2==seqs.len[s2] ? scrs.endgapcost : scrs.gapcost;

  ALLOC(Vf,int,len1+1);
  ALLOC(Hf,int,len1+1);
  ALLOC(b,int,len1+1);
  ALLOC(Vb,int,len1+1);
  ALLOC(Hb,int,len1+1);

  /* calculate forward part of matrix */
  f[0] = 0;
  Vf[0] = lendgapcost2; /* INFITITY (should not be set to INT_MAX !!!) */
  Hf[0] = lendgapcost1; /* INFITITY */
  for(i=1; i<=len1; ++i) {
    f[i] = Vf[i] = Vf[i-1]+scrs.scr[t1[i-1]][GAP];
    Hf[i] = f[i]+(i==len1?rendgapcost1:scrs.gapcost); /* INFINITY */
  }
  for(j=1; j<=c-l2; ++j) {
    mem = f[0];
    f[0] = Hf[0] += scrs.scr[GAP][t2[j-1]];
    Vf[0] = f[0]+(j==len2?rendgapcost2:scrs.gapcost); /* INFINITY */
    for(i=1; i<=len1; ++i) {
      D = mem+scrs.scr[t1[i-1]][t2[j-1]];
      Vf[i] = MIN(f[i-1]+(j==len2?rendgapcost2:scrs.gapcost),Vf[i-1])
              +scrs.scr[t1[i-1]][GAP];
      Hf[i] = MIN(f[i]+(i==len1?rendgapcost1:scrs.gapcost),Hf[i])
              +scrs.scr[GAP][t2[j-1]];
      mem = f[i];
      f[i] = MIN3(D,Vf[i],Hf[i]);
    }
  }

  /* calculate reverse part of matrix */
  b[len1] = 0;
  Vb[len1] = rendgapcost2; /* INFINITY */
  Hb[len1] = rendgapcost1; /* INFINITY */
  for(i=len1-1; i>=0; --i) {
    b[i] = Vb[i] = Vb[i+1]+scrs.scr[t1[i]][GAP];
    Hb[i] = b[i]+(i==0?lendgapcost1:scrs.gapcost); /* INFINITY */
  }
  for(j=len2-1; j>=c-l2; --j) {
    mem = b[len1];
    b[len1] = Hb[len1] += scrs.scr[GAP][t2[j]];
    Vb[len1] = b[len1]+(j==0?lendgapcost2:scrs.gapcost); /* INFINITY */
    for(i=len1-1; i>=0; --i) {
      D = mem+scrs.scr[t1[i]][t2[j]];
      Vb[i] = MIN(b[i+1]+(j==0?lendgapcost2:scrs.gapcost),Vb[i+1])
              +scrs.scr[t1[i]][GAP];
      Hb[i] = MIN(b[i]+(i==0?lendgapcost1:scrs.gapcost),Hb[i])
              +scrs.scr[GAP][t2[j]];
      mem = b[i];
      b[i] = MIN3(D,Vb[i],Hb[i]);
    }
  }

  /* calculate optimal alignment score and column of additional cost matrix */
  score = INT_MAX;
  for(i=0; i<=len1; ++i) {
    f[i] = MIN3(f[i]+b[i],Vf[i]+Vb[i]-(c==0 ? lendgapcost2 : 
                                   (c==len2 ? rendgapcost2 : scrs.gapcost)),
                          Hf[i]+Hb[i]-(i==0 ? lendgapcost1 :
                                   (i==len1 ? rendgapcost1 : scrs.gapcost)));
    score = MIN(score,f[i]);
  }
  for(i=0; i<=len1; ++i)
    f[i] -= score;

  FREE(Vf);
  FREE(Hf);
  FREE(b);
  FREE(Vb);
  FREE(Hb);
  return(score);
} /* calc_column() */

/*---------------------------------------------------------------------------*/
void calc_matrix(int s1,int s2, int l1,int l2, int u1,int u2,
                 int lo1,int lo2, int uo1,int uo2, matrix m, int score)
{
  extern sequences seqs;
  extern scores scrs;

  sequence t1,t2;
  int len1,len2, olen1,olen2, i,j, mem,D,
      lendgapcost1,rendgapcost1,lendgapcost2,rendgapcost2;
  column f,Vf,Hf, b,Vb,Hb;
  matrix Vm,Hm;

  t1 = seqs.s[s1]+l1;
  t2 = seqs.s[s2]+l2;
  len1 = u1-l1;
  len2 = u2-l2;
  olen1 = uo1-lo1;
  olen2 = uo2-lo2;

  lendgapcost1 = l1==0 ? scrs.endgapcost : scrs.gapcost;
  rendgapcost1 = u1==seqs.len[s1] ? scrs.endgapcost : scrs.gapcost;
  lendgapcost2 = l2==0 ? scrs.endgapcost : scrs.gapcost;
  rendgapcost2 = u2==seqs.len[s2] ? scrs.endgapcost : scrs.gapcost;

  ALLOC(f,int,uo1+1);
  ALLOC(Vf,int,uo1+1);
  ALLOC(Hf,int,uo1+1);
  ALLOC(b,int,len1+1); /* b[0] ... b[lo1-1] never used */
  ALLOC(Vb,int,len1+1); /* Vb[0] ... Vb[lo1-1] never used */
  ALLOC(Hb,int,len1+1); /* Hb[0] ... Hb[lo1-1] never used */

  ALLOC(Vm,int*,olen1+1);
  ALLOC(Hm,int*,olen1+1);
  for(i=0; i<=olen1; ++i) {
    ALLOC(Vm[i],int,olen2+1);
    ALLOC(Hm[i],int,olen2+1);
  }

  /* calculate forward matrices */
  f[0] = 0;
  Vf[0] = lendgapcost2; /* INFITITY */
  Hf[0] = lendgapcost1; /* INFITITY */
  for(i=1; i<=uo1; ++i) {
    f[i] = Vf[i] = Vf[i-1]+scrs.scr[t1[i-1]][GAP];
    Hf[i] = f[i]+(i==len1?rendgapcost1:scrs.gapcost); /* INFINITY */
  }
  for(j=1; j<=lo2; ++j) {
    mem = f[0];
    f[0] = Hf[0] += scrs.scr[GAP][t2[j-1]];
    Vf[0] = f[0]+(j==len2?rendgapcost2:scrs.gapcost); /* INFINITY */
    for(i=1; i<=uo1; ++i) {
      D = mem+scrs.scr[t1[i-1]][t2[j-1]];
      Vf[i] = MIN(f[i-1]+(j==len2?rendgapcost2:scrs.gapcost),Vf[i-1])
              +scrs.scr[t1[i-1]][GAP];
      Hf[i] = MIN(f[i]+(i==len1?rendgapcost1:scrs.gapcost),Hf[i])
              +scrs.scr[GAP][t2[j-1]];
      mem = f[i];
      f[i] = MIN3(D,Vf[i],Hf[i]);
    }
  }

  /* copy column lo2 into matrices */
  for(i=lo1; i<=uo1; ++i) {
    m[i-lo1][0] = f[i];
    Hm[i-lo1][0] = Hf[i];
    Vm[i-lo1][0] = Vf[i];
  }

  /* calculate rest */
  for(/* j unchanged */; j<=uo2; ++j) {
    mem = f[0];
    f[0] = Hf[0] += scrs.scr[GAP][t2[j-1]];
    Vf[0] = f[0]+(j==len2?rendgapcost2:scrs.gapcost); /* INFINITY */
    for(i=1; i<=lo1; ++i) {
      D = mem+scrs.scr[t1[i-1]][t2[j-1]];
      Vf[i] = MIN(f[i-1]+(j==len2?rendgapcost2:scrs.gapcost),Vf[i-1])
              +scrs.scr[t1[i-1]][GAP];
      Hf[i] = MIN(f[i]+(i==len1?rendgapcost1:scrs.gapcost),Hf[i])
              +scrs.scr[GAP][t2[j-1]];
      mem = f[i];
      f[i] = MIN3(D,Vf[i],Hf[i]);
    }
    /* copy entry of lo1-th row into matrices */
    m[0][j-lo2] = f[lo1];
    Hm[0][j-lo2] = Hf[lo1];
    Vm[0][j-lo2] = Vf[lo1];

    /* calculate rest inside matrices */
    for(/* i unchanged */; i<=uo1; ++i) {
      D = m[i-1-lo1][j-1-lo2]+scrs.scr[t1[i-1]][t2[j-1]];
      Vm[i-lo1][j-lo2] = MIN(m[i-1-lo1][j-lo2]
                             +(j==len2?rendgapcost2:scrs.gapcost),
                             Vm[i-1-lo1][j-lo2])+scrs.scr[t1[i-1]][GAP];
      Hm[i-lo1][j-lo2] = MIN(m[i-lo1][j-1-lo2]
                             +(i==len1?rendgapcost1:scrs.gapcost),
                             Hm[i-lo1][j-1-lo2])+scrs.scr[GAP][t2[j-1]];;
      m[i-lo1][j-lo2] = MIN3(D,Vm[i-lo1][j-lo2],Hm[i-lo1][j-lo2]);
    }
  }

  /* calculate reverse matrices and on the fly additional cost matrix */
  b[len1] = 0;
  Vb[len1] = rendgapcost2; /* INFINITY */
  Hb[len1] = rendgapcost1; /* INFINTY */
  for(i=len1-1; i>=lo1; --i) {
    b[i] = Vb[i] = Vb[i+1]+scrs.scr[t1[i]][GAP];
    Hb[i] = b[i]+(i==0?lendgapcost1:scrs.gapcost); /* INFINITY */
  }
  for(j=len2-1; j>=uo2; --j) {
    mem = b[len1];
    b[len1] = Hb[len1] += scrs.scr[GAP][t2[j]];
    Vb[len1] = b[len1]+(j==0?lendgapcost2:scrs.gapcost); /* INFINITY */
    for(i=len1-1; i>=lo1; --i) {
      D = mem+scrs.scr[t1[i]][t2[j]];
      Vb[i] = MIN(b[i+1]+(j==0?lendgapcost2:scrs.gapcost),Vb[i+1])
              +scrs.scr[t1[i]][GAP];
      Hb[i] = MIN(b[i]+(i==0?lendgapcost1:scrs.gapcost),Hb[i])
              +scrs.scr[GAP][t2[j]];
      mem = b[i];
      b[i] = MIN3(D,Vb[i],Hb[i]);
    }
  }
  /* calculate last column of additional cost matrix */
  for(i=uo1; i>=lo1; --i)
    m[i-lo1][olen2] = MIN3(m[i-lo1][olen2]+b[i],
                     Vm[i-lo1][olen2]+Vb[i]-(uo2==len1?rendgapcost2:
                                            (uo2==0?lendgapcost2:scrs.gapcost)),
                     Hm[i-lo1][olen2]+Hb[i]-(i==len1?rendgapcost1:
                                            (i==0?lendgapcost1:scrs.gapcost)))
                      -score;

  /* calculate additional cost matrix on the fly */
  for(/* j unchanged */; j>=lo2; --j) { 
    mem = b[len1];
    b[len1] = Hb[len1] += scrs.scr[GAP][t2[j]];
    Vb[len1] = b[len1]+(j==0?lendgapcost2:scrs.gapcost); /* INFINITY */
    for(i=len1-1; i>=uo1; --i) {
      D = mem+scrs.scr[t1[i]][t2[j]];
      Vb[i] = MIN(b[i+1]+(j==0?lendgapcost2:scrs.gapcost),Vb[i+1])
              +scrs.scr[t1[i]][GAP];
      Hb[i] = MIN(b[i]+(i==0?lendgapcost1:scrs.gapcost),Hb[i])
              +scrs.scr[GAP][t2[j]];
      mem = b[i];
      b[i] = MIN3(D,Vb[i],Hb[i]);
    }
    /* entry of last row */
    m[olen1][j-lo2] = MIN3(m[olen1][j-lo2]+b[uo1],
                   Vm[olen1][j-lo2]+Vb[uo1]-(j==0?lendgapcost2:scrs.gapcost),
                   Hm[olen1][j-lo2]+Hb[uo1]-(uo1==len1?rendgapcost1:
                                            (uo1==0?lendgapcost1:scrs.gapcost)))
                      -score;

    for(/* i unchanged */; i>=lo1; --i) {
      D = mem+scrs.scr[t1[i]][t2[j]];
      Vb[i] = MIN(b[i+1]+(j==0?lendgapcost2:scrs.gapcost),Vb[i+1])
              +scrs.scr[t1[i]][GAP];
      Hb[i] = MIN(b[i]+(i==0?lendgapcost1:scrs.gapcost),Hb[i])
              +scrs.scr[GAP][t2[j]];
      mem = b[i];
      b[i] = MIN3(D,Vb[i],Hb[i]);
      m[i-lo1][j-lo2] = MIN3(m[i-lo1][j-lo2]+b[i],
                 Vm[i-lo1][j-lo2]+Vb[i]-(j==0?lendgapcost2:scrs.gapcost),
                 Hm[i-lo1][j-lo2]+Hb[i]-(i==0?lendgapcost1:scrs.gapcost))-score;
    }
  }

  FREE(f);
  FREE(Vf);
  FREE(Hf);
  FREE(b);
  FREE(Vb);
  FREE(Hb);

  for(i=0; i<=olen1; ++i) {
    FREE(Vm[i]);
    FREE(Hm[i]);
  }
  FREE(Vm);
  FREE(Hm);
} /* calc_matrix() */

/****** EOF (dypr.c) *********************************************************/
