/*****************************************************************************
 *  FILE   : dypr.h
 *  AUTHOR : Jens Stoye
 *  DATE   : 05/1997
 *  RCS-ID : $Id: dypr.h,v 1.1 1997/07/11 11:15:47 jones Exp jones $
 *****************************************************************************/
#ifndef DYPR_H
#define DYPR_H

/***************************************************************************** 
 * INCLUDES
 *****************************************************************************/
#include "global.h" /* data types */

/***************************************************************************** 
 * FUNCTION PROTOTYPES                                                         
 *****************************************************************************/

/*===========================================================================*/
/* -- CALC_COLUMN --
 * Calculate 'c'-th column of the additional cost matrix of sequences no.
 * 's1' and 's2' from global 'seqs' between 'l1' and 'u1' ('l2' and 'u2', 
 * resp.) and return the score of the optimal alignment according to global
 * distance score 'scrs'.
 * Linear gap costs of the form g(k) = scrs.gapcost + \sum_i scrs.scr[i][GAP].
 * Terminal gaps (i.e. l1==0, u1==seqs.len[s1], l2==0, or u2==seqs.len[s2])
 * are charged with 'scrs.endgapcost' */
/*---------------------------------------------------------------------------*/
int calc_column(
  int s1, /* number of first sequence */
  int s2, /* number of second sequence */
  int l1, /* lower bound of first sequence */
  int l2, /* lower bound of second sequence */
  int u1, /* upper bound of first sequence */
  int u2, /* upper bound of second sequence */
  int c, /* index of column to be calculated */
  column f /* column */
);

/*===========================================================================*/
/* -- CALC_MATRIX --
 * Calculate additional cost matrix of sequences no. 's1' and 's2' from global
 * seqs between 'l1' and 'u1' ('l2' and 'u2', resp.) according to global
 * distance score 'scrs'. Save only the relevant part between 'lo1' and 'uo1'
 * ('lo2' and 'uo2', resp.). 'score' is the optimal alignment score of 's1'
 * and 's2' (must be determined previously, e.g. using 'calc_column').
 * Linear gap costs of the form g(k) = scrs.gapcost + \sum_i scrs.scr[i][GAP] .
 * Terminal gaps (i.e. l1==0, u1==seqs.len[s1], l2==0, or u2==seqs.len[s2])
 * are charged with 'scrs.endgapcost' */
/*---------------------------------------------------------------------------*/
void calc_matrix(
  int s1, /* number of first sequence */
  int s2, /* number of second sequence */
  int l1, /* lower bound of first sequence */
  int l2, /* lower bound of second sequence */
  int u1, /* upper bound of first sequence */
  int u2, /* upper bound of second sequence */
  int lo1, /* lower offset of relevant region in first sequence */
  int lo2, /* lower offset of relevant region in second sequence */
  int uo1, /* upper offset of relevant region in first sequence */
  int uo2, /* upper offset of relevant region in second sequence */
  matrix m, /* matrix */
  int score /* optimal score */
);

#endif /* DYPR_H */
/****** EOF (dypr.h) *********************************************************/
