/*****************************************************************************
 *  FILE   : global.h
 *  AUTHOR : Jens Stoye
 *  DATE   : 05/1997
 *  RCS-ID : $Id: global.h,v 1.6 1997/07/31 16:18:56 jones Exp jones $
 *****************************************************************************/
#ifndef GLOBAL_H
#define GLOBAL_H

/*****************************************************************************
 * INCLUDES
 *****************************************************************************/
#include <stdlib.h> /* malloc(),realloc(),free(),sizeof(),exit() */
#include <stdio.h> /* fopen(),fprintf(),fscanf(),fgets(),fclose(),feof() */

/* ++pac */
#ifdef _PLAN9
#include <sys/types.h> /* ++pac */
#endif

#include <sys/stat.h> /* remove() */
#include <limits.h> /* CLK_TCK */
#include <unistd.h> /* CLK_TCK->sysconf() */
#include <time.h> /* CLK_TCK */

/*****************************************************************************
 * DEFAULT VALUES
 *****************************************************************************/
#define DEFAULT_Q FALSE
#define DEFAULT_FREE_SHIFT TRUE
#define DEFAULT_A FALSE
#define DEFAULT_O TRUE
#define DEFAULT_WEIGHT_INTENSITY 0.0
#define DEFAULT_WEIGHTS_FILENAME NULL
#define DEFAULT_T FALSE
#define DEFAULT_SCRS_PROTEIN_FILENAME "blosum62"
#define DEFAULT_SCRS_IUPAC_FILENAME "dnarna"
#define DEFAULT_SCRS_RNASS_FILENAME "rnasec"
#define DEFAULT_SCRS_UNKNOWN_FILENAME "blosum62"
#define DEFAULT_L 30
#define DEFAULT_W 0
#define DEFAULT_FORMAT PIR

/*****************************************************************************
 * GLOBAL CONSTANTS AND MACROS
 *****************************************************************************/
#define MAX_LENGTH  256

#define USAGE "[-q] [-g] [-a] [-o] [-b (B|wf_file)] [-t [order]] [-c cost_file] [-l L] [-w W] [-f F] <sequences>"
#define RNASS_USAGE "[-q] [-g] [-a] [-o] [-b (B|wf_file)] [-t [order]] [-c cost_file] [-l L] [-w W] [-f F] <sequences> <structures>"

#define ALN_SUFFIX "dca"
#define PIR_SUFFIX "dca"
#define NEX_SUFFIX "nex"
#define DCA_SUFFIX "dca"

#define ALN 1
#define PIR 2
#define NEX 3
#define DCA 4

/*-- scores -----------------------------------------------------------------*/
#define SCORES_DIR "DCA_COST_DIR" /* environment variable */
#define SCORES_COMMENT "#"
#define SCORES_ABC 25 /* 26 is not possible, since 'Z'-26=0==GAP */
#define GAP 0
#define SCORES_LINE_DISTANCE 5

#define PROTEIN_TYPE 1
#define PROTEIN_NAME "protein"
#define PROTEIN_TAG '%'
#define PROTEIN_NUM 20
/* list of amino acids: -  C  S  T  P  A  G  N  D  E  Q  H  R  K  M  I  L  V  F  Y  W */
#define PROTEIN_IDC   { 0,23, 7, 6,10,25,19,12,22,21, 9,18, 8,15,13,17,14, 4,20, 1, 3}

#define IUPAC_TYPE 2
#define IUPAC_NAME "dna/rna (iupac)"
#define IUPAC_TAG '#'
#define IUPAC_NUM 16
/* list of bases:   -  A  C  G  T  C  R  Y  K  M  S  W  B  D  H  V  N */
#define IUPAC_IDC { 0,25,23,19, 6, 5, 8, 1,15,13, 7, 3,24,22,18, 4,12}

#define DNA_TYPE 3
#define DNA_NAME "dna"
#define DNA_TAG '#'
#define DNA_NUM 4
/* list of bases: -  A  C  G  T */
#define DNA_IDC { 0,25,23,19, 6}

#define RNA_TYPE 4
#define RNA_NAME "rna"
#define RNA_TAG '#'
#define RNA_NUM 4
/* list of bases: -  A  C  G  U */
#define RNA_IDC { 0,25,23,19, 5}
#define IS_RNA(C) ((int)(C)=='-' || (int)(C)=='A' || (int)(C)=='C' || \
                   (int)(C)=='G' || (int)(C)=='U')

#define RNASS_TYPE 5
#define RNASS_NAME "rna secondary structure"
#define RNASS_TAG '$'
#define RNASS_NUM 12
/* ss elements:     -  A  B  X  C  D  E  G  H  I  U  V  W  */
#define RNASS_IDC { 0,25,24, 2,23,22,21,19,18,17, 5, 4, 3}
#define UNKNOWN_TYPE 0
#define UNKNOWN_NAME "unknown"
#define UNKNOWN_TAG '>'
#define UNKNOWN_NUM SCORES_ABC /* 25 */
/* list of letters:  -  A  B  C  D  E  F  G  H  I  J  K  L  M  N  O  P  Q  R  S  T  U  V  W  X  Y */
#define UNKNOWN_IDC {0,25,24,23,22,21,20,19,18,17,16,15,14,13,12,11,10, 9, 8, 7, 6, 5, 4, 3, 2, 1}

/*-- seqs -------------------------------------------------------------------*/
#define SEQS_DIR "DCA_SEQS_DIR" /* environment variable */
#define SEQS_OUTPUT_LINE_LEN 60
#define NAME_SPACER '_'
#define NORMALIZE_FACTOR 100 /* for length normalization of alignment scores */

#define TYPE2TAG(T) (T==PROTEIN_TYPE ? PROTEIN_TAG : \
                    (T==IUPAC_TYPE ? IUPAC_TAG : \
                    (T==DNA_TYPE ? DNA_TAG : \
                    (T==RNA_TYPE ? RNA_TAG : \
                    (T==RNASS_TYPE ? RNASS_TAG : UNKNOWN_TAG)))))
#define TYPE2NAME(T) (T==PROTEIN_TYPE ? PROTEIN_NAME : \
                     (T==IUPAC_TYPE ? IUPAC_NAME : \
                     (T==DNA_TYPE ? DNA_NAME : \
                     (T==RNA_TYPE ? RNA_NAME : \
                     (T==RNASS_TYPE ? RNASS_NAME : UNKNOWN_NAME)))))
#define TAG2TYPE(T) (T==PROTEIN_TAG ? PROTEIN_TYPE : \
                    (T==IUPAC_TAG ? IUPAC_TYPE : \
                    (T==DNA_TAG ? DNA_TYPE : \
                    (T==RNA_TAG ? RNA_TYPE : \
                    (T==RNASS_TAG ? RNASS_TYPE : UNKNOWN_TYPE)))))

#define SEQS_GAP_TAG '-'
#define SEQS_ERROR_TAG '*'
#define SEQS_ALLOCSTEP 1000

/*-- code -------------------------------------------------------------------*/
#define CODE_NEW_SEQUENCE -1
#define CODE_ERROR -2

/*-- level ------------------------------------------------------------------*/
#define NO_LEVEL -1
#define ERROR_LEVEL -2
#define DECODE_LEVEL(L) ((L)==NO_LEVEL ? ' ' :\
                         ((L)==ERROR_LEVEL ? '.' : '0'+(L)%10))

/*-- msa --------------------------------------------------------------------*/
#define MSA_OPTIMAL_TEXT      "***  Optimal Multiple Alignment  ***"
#define MSA_HEURISTIC_TEXT    "***  Heuristic Multiple Alignment  ***"
#define MSA_END_GAPS_TEXT     "End gaps not penalized."
#define MSA_TIME_TEXT         "Elapsed time ="
#define MSA_OPTIMAL_LEVEL     NO_LEVEL
#define MSA_HEURISTIC_LEVEL   ERROR_LEVEL

#ifndef MSA_CALL /* might be pre-defined in compiler command line */
#define MSA_CALL "msa"
#endif
#define MSA_INPUT_FILENAME "/tmp/msa_i"
#define MSA_OUTPUT_FILENAME "/tmp/msa_o"
#define MSA_COST_FILENAME "/tmp/msa_c"
#define MSA_FORCE_FILENAME "/tmp/msa_f"

/*-- splits -----------------------------------------------------------------*/
#ifndef SPLITS_CALL /* might be pre-defined in compiler command line */
#define SPLITS_CALL "splits"
#endif
#define SPLITS_INPUT_FILENAME "/tmp/splits_i"
#define SPLITS_OUTPUT_FILENAME "/tmp/splits_o"

/*-- times ------------------------------------------------------------------*/
#ifndef CLK_TCK
#define CLK_TCK 60.0
#endif

/*-- bool -------------------------------------------------------------------*/
#define BOOL unsigned int
#define FALSE 0
#define TRUE  1

/*-- minmax -----------------------------------------------------------------*/
#define MAX(X,Y) (((X) > (Y)) ? (X) : (Y))
#define MIN(X,Y) (((X) < (Y)) ? (X) : (Y))
#define MIN3(X,Y,Z) (((X) < (Y)) ? ((X) < (Z) ? (X) : (Z))\
                                 : ((Y) < (Z) ? (Y) : (Z)))
#define ABS(X) (((X) < 0) ? -(X) : (X))
#define ROUND(X) ((int)(X+0.5))

/*-- alloc ------------------------------------------------------------------*/
#define ALLOC(P,T,S) if((P=(T*)malloc(sizeof(T)*(S))) == NULL) {\
                       fprintf(stderr,"malloc failed.\n");\
                       exit(1);\
                     }
#define REALLOC(P,T,S) if((P=(T*)realloc(P,sizeof(T)*(S))) == NULL) {\
                         fprintf(stderr,"realloc failed.\n");\
                         exit(1);\
                       }
#define FREE(P) free(P)

/*-- files ------------------------------------------------------------------*/
#define FOPEN_READABLE(P,N) if((P=fopen(N,"r")) == NULL)\
                              fprintf(stderr,"Can't open file \"%s\".\n",N)
#define FOPEN_READABLE_BASE(P,N,B) if((P=fopen(N,"r")) == NULL)\
                                     if((t1=getenv(B)) == NULL)\
                                       fprintf(stderr,\
                                               "Can't open file \"%s\".\n",N);\
                                     else {\
                                       sprintf(t2,"%s/%s",t1,N);\
                                       if((P=fopen(t2,"r")) == NULL)\
                                         fprintf(stderr,\
                                               "Can't open file \"%s\".\n",N);\
                                     }
#define FOPEN_WRITEABLE(P,N) if((P=fopen(N,"w")) == NULL)\
                               fprintf(stderr,"Can't open file \"%s\".\n",N)
#define FCLOSE(P) fclose(P)

#define FGETS(P,L,F) if(feof(F))\
                       fprintf(stderr,"fgets failed.\n");\
                     else {\
                       fgets(P,L,F);\
                       P[strlen(P)-1] = '\0'; /* remove terminal newline */\
                     }

#define FSCANF(P,S,V) if(feof(P))\
                        fprintf(stderr,"fscanf failed.\n");\
                      else fscanf(P,S,V)

#define REMOVE(N) if(remove(N) != 0)\
                    fprintf(stderr,"Can't remove file \"%s\".\n",N)

/*****************************************************************************
 * DATA TYPES
 *****************************************************************************/
struct PARAMETERS {
  BOOL q; /* quiet */
  BOOL free_shift; /* do not penalize terminal gaps */
  BOOL a; /* use approximate cut positions */
  BOOL o; /* send information about progress of the program to stderr */
  float weight_intensity; /* weight intensity */
  char weights_filename[MAX_LENGTH]; /* weights file name */
  BOOL spantree; /* tree-based sequence order */
  char scrs_filename[MAX_LENGTH]; /* score matrix file name */
  int l; /* recursion stop size L */
  int w; /* re-align window size W */
  int format; /* output format */
  char seqs_filename[MAX_LENGTH]; /* sequences file name */
  char seqs_filename_base[MAX_LENGTH]; /* filename base */
  char aln_filename[MAX_LENGTH]; /* aln filename */
  char pir_filename[MAX_LENGTH]; /* pir filename */
  char nex_filename[MAX_LENGTH]; /* nexus filename */
  char dca_filename[MAX_LENGTH]; /* dca filename */
};
typedef struct PARAMETERS parameters;

/*-- seqs -------------------------------------------------------------------*/
typedef int letter; /* letters are encoded as integers */
typedef letter* sequence;

struct SEQUENCES {
  int num;       /* number of sequences */
  int type;      /* sequence type */
  char **name;   /* vector of sequence names */
  int *len;      /* vector of sequence lengths */
  sequence *s;   /* matrix of sequence entries */
  int **pwscore; /* symmetric matrix of normalized pairwise optimal scores */
  float **wf;    /* symmetric matrix of pairwise weight factors */
  int *slist;    /* sequence order when tree-based */
};
typedef struct SEQUENCES sequences;

struct ALIGNMENT {
  int num;       /* number of sequences in alignment */
  int type;      /* sequence type */
  char **name;   /* vector of sequence names */
  int len;       /* alignment length */
  sequence *a;   /* matrix of alignment entries */
  BOOL msa_approx; /* flag if approximate alignment */
};
typedef struct ALIGNMENT alignment;

/*-- scores -----------------------------------------------------------------*/
struct SCORES { /* general score matrix type */
  int num; /* number of letters */
  int type; /* score matrix type */
  int scr[SCORES_ABC+1][SCORES_ABC+1]; /* score matrix (+1 for gap) */
  int min; /* minimum value of matrix entries */
  int max; /* maximum value of matrix entries */ 
  int gapcost; /* (additional) cost of an internal gap */
  int endgapcost; /* (additional) cost of a terminal gap */
  letter idc[SCORES_ABC+1]; /* index counter for matrix output */
};
typedef struct SCORES scores;

/*-- dypr -------------------------------------------------------------------*/
typedef int* tuple;
typedef int* column;
typedef column* matrix;

#endif /* GLOBAL_H */
/****** EOF (global.h) *******************************************************/

