/*****************************************************************************
 *  FILE   : scores.c
 *  AUTHOR : Jens Stoye
 *  DATE   : 05/1997
 *  RCS-ID : $Id: scores.c,v 1.2 1997/07/25 12:31:44 jones Exp jones $
 *****************************************************************************/

/*****************************************************************************
 * INCLUDES
 *****************************************************************************/
#include <stdio.h> /* FILE,fprintf() */
#include <string.h> /* strdup() */
#include <ctype.h> /* islower(),toupper(),isalpha() */

#include "global.h" /* data types,BOOL,TRUE,FALSE,MIN(),MAX(),ALLOC() */
#include "scores.h" /* function prototypes */

/*****************************************************************************
 * FUNCTION DEFINITIONS
 *****************************************************************************/
letter encode(int c)
{
  if(islower(c))
    c = toupper(c);
  if(isalpha(c)) /* letter (amino acid, base) */
    return((int)'Z'-c);
  if(c == (int)SEQS_GAP_TAG) /* gap */
    return(GAP);
  if((c==(int)PROTEIN_TAG) ||
     (c==(int)IUPAC_TAG) ||
     (c==(int)DNA_TAG) ||
     (c==(int)RNA_TAG) ||
     (c==(int)RNASS_TAG) ||
     (c==(int)UNKNOWN_TAG)) /* comment line */
    return(CODE_NEW_SEQUENCE);
  return(CODE_ERROR); /* anything else: error */
} /* encode() */

/*---------------------------------------------------------------------------*/
char decode(letter i)
{
  return(i<0 ? SEQS_ERROR_TAG : i==GAP ? SEQS_GAP_TAG : 'Z'-i);
} /* decode() */

/*---------------------------------------------------------------------------*/
void read_scores(FILE *fptr, scores *sc)
{
  char a,b;
  int i,j, value;
  float fvalue;
  letter protein_idc[PROTEIN_NUM+1]=PROTEIN_IDC;
  letter iupac_idc[IUPAC_NUM+1]=IUPAC_IDC;
  letter dna_idc[DNA_NUM+1]=DNA_IDC;
  letter rna_idc[RNA_NUM+1]=RNA_IDC;
  letter rnass_idc[RNASS_NUM+1]=RNASS_IDC;
  letter unknown_idc[UNKNOWN_NUM+1]=UNKNOWN_IDC;
  BOOL exists[SCORES_ABC+1];

  /* initialize exists array */  
  for(i=0; i<SCORES_ABC; ++i)
    exists[i] = FALSE;

  /* initialize score matrix with zero */
  for(i=0; i<=SCORES_ABC; ++i)
    for(j=0; j<=SCORES_ABC; ++j)
      sc->scr[i][j] = 0;
  sc->min = 0;
  sc->max = 0;

  /* read matrix entries from fptr */
  FSCANF(fptr," %f\n",&fvalue);
  sc->gapcost = ROUND(fvalue);
  while(fscanf(fptr," %c %c %f\n",&a,&b,&fvalue)==3) {
    value = ROUND(fvalue);
    i = encode(a);
    j = encode(b);
    sc->scr[i][j] = sc->scr[j][i] = value;
    sc->min = MIN(sc->min,value);
    sc->max = MAX(sc->max,value);
    exists[i] = exists[j] = TRUE;
  }

  /* find out score matrix type */
  for(sc->num=0,i=0; i<SCORES_ABC; ++i) /* do not count gap letter */
    if(exists[i])
      ++sc->num;

  if(sc->num==PROTEIN_NUM && exists[encode('-')] &&
     exists[encode('A')] && exists[encode('C')] &&
     exists[encode('D')] && exists[encode('E')] &&
     exists[encode('F')] && exists[encode('G')] &&
     exists[encode('H')] && exists[encode('I')] &&
     exists[encode('K')] && exists[encode('L')] &&
     exists[encode('M')] && exists[encode('N')] &&
     exists[encode('P')] && exists[encode('Q')] &&
     exists[encode('R')] && exists[encode('S')] &&
     exists[encode('T')] && exists[encode('V')] &&
     exists[encode('W')] && exists[encode('Y')]) {
    sc->type = PROTEIN_TYPE;
    for(i=0; i<=sc->num; ++i)
      sc->idc[i] = protein_idc[i];
  }
  else
    if(sc->num==IUPAC_NUM && exists[encode('-')] &&
       exists[encode('A')] && exists[encode('C')] &&
       exists[encode('G')] && exists[encode('T')] &&
       exists[encode('U')] && exists[encode('R')] &&
       exists[encode('Y')] && exists[encode('K')] &&
       exists[encode('M')] && exists[encode('S')] &&
       exists[encode('W')] && exists[encode('B')] &&
       exists[encode('D')] && exists[encode('H')] &&
       exists[encode('V')] && exists[encode('N')]) {
      sc->type = IUPAC_TYPE;
      for(i=0; i<=sc->num; ++i)
        sc->idc[i] = iupac_idc[i];
    }
    else
      if(sc->num==DNA_NUM && exists[encode('-')] &&
         exists[encode('A')] && exists[encode('C')] &&
         exists[encode('G')] && exists[encode('T')]) {
        sc->type = DNA_TYPE;
        for(i=0; i<=sc->num; ++i)
          sc->idc[i] = dna_idc[i];
      }
      else
        if(sc->num==RNA_NUM && exists[encode('-')] &&
           exists[encode('A')] && exists[encode('C')] &&
           exists[encode('G')] && exists[encode('U')]) {
          sc->type = RNA_TYPE;
          for(i=0; i<=sc->num; ++i)
            sc->idc[i] = rna_idc[i];
        }
        else
          if(sc->num==RNASS_NUM && exists[encode('-')] &&
             exists[encode('A')] && exists[encode('B')] &&
             exists[encode('X')] && exists[encode('C')] &&
             exists[encode('D')] && exists[encode('E')] &&
             exists[encode('G')] && exists[encode('H')] &&
             exists[encode('I')] && exists[encode('U')] &&
             exists[encode('V')] && exists[encode('W')]) {
            sc->type = RNASS_TYPE;
            for(i=0; i<=sc->num; ++i)
              sc->idc[i] = rnass_idc[i];
          }
          else {
            sc->num = UNKNOWN_NUM;
            sc->type = UNKNOWN_TYPE;
            for(i=0; i<=sc->num; ++i)
              sc->idc[i] = unknown_idc[i];
          }
  
} /* read_scores() */

/*---------------------------------------------------------------------------*/
void write_scores(FILE *fptr, scores *sc)
{
  int i,j,k;

  fprintf(fptr,"\n%s",SCORES_COMMENT);
  fprintf(fptr," %s score matrix:",TYPE2NAME(sc->type));
  fprintf(fptr," %d entries",sc->num);

  fprintf(fptr,"\n%s",SCORES_COMMENT);
  fprintf(fptr," gapscore=%d",sc->gapcost);
  fprintf(fptr,", terminal gapscore=%d",sc->endgapcost);

  fprintf(fptr,"\n%s",SCORES_COMMENT);
  fprintf(fptr," min=%d",sc->min);
  fprintf(fptr,", max=%d",sc->max);
  fprintf(fptr,"\n\n");

  for (i=0; i<=sc->num; ++i) {
    fprintf(fptr,"%c ",decode(sc->idc[i]));
    for (j=0; j<=i; ++j)
      fprintf(fptr," % 3d",sc->scr[sc->idc[i]][sc->idc[j]]);
    fprintf(fptr, "\n");
    if(i%SCORES_LINE_DISTANCE == SCORES_LINE_DISTANCE-1) {
      for(k=0; k<2+4*(i+2); ++k)
        fputc('-',fptr);
      fputc('\n',fptr);
    }
  }
  fprintf(fptr,"  ");
  for (i=0; i<=sc->num; ++i)
    fprintf(fptr,"   %c",decode(sc->idc[i]));
  fprintf(fptr,"\n\n");
} /* write_scores() */

/*---------------------------------------------------------------------------*/
void write_scores_msa(FILE *fptr, scores *sc)
{
  int i,j;

  fprintf(fptr,"\n%d\n",sc->gapcost);

  for(i=0; i<=sc->num; ++i)
     for(j=i; j<=sc->num; ++j)
       fprintf(fptr,"%c %c %d\n",decode(sc->idc[j]),decode(sc->idc[i]),
                                 sc->scr[sc->idc[j]][sc->idc[i]]);
  fprintf(fptr,"\n");
} /* write_scores_msa() */

/*---------------------------------------------------------------------------*/
void reset_gapcost(scores *sc, BOOL free_shift)
{
  sc->endgapcost = free_shift ? 0 : sc->gapcost;
} /* reset_gapcost() */

/****** EOF (scores.c) *******************************************************/
