/*****************************************************************************
 *  FILE   : scores.h
 *  AUTHOR : Jens Stoye
 *  DATE   : 05/1996
 *  RCS-ID : $Id: scores.h,v 1.1 1997/07/11 11:15:47 jones Exp jones $
 *****************************************************************************/
#ifndef SCORES_H
#define SCORES_H

/*****************************************************************************
 * INCLUDES
 *****************************************************************************/
#include <stdio.h> /* FILE */

#include "global.h" /* data types,BOOL */

/*****************************************************************************
 * FUNCTION PROTOTYPES
 *****************************************************************************/

/*===========================================================================*/
/* -- ENCODE --
 * Encode the characters of an alphabet as 'letter's:
 * '-' --> GAP=0, 'A' --> 26, 'B' --> 25, ... , 'Y' --> 1, 'Z' --> 0=GAP */
/*---------------------------------------------------------------------------*/
letter encode(
  int c /* character to be encoded */
);

/*===========================================================================*/
/* -- DECODE --
 * Decode letter 'i' back to characters:
 * GAP=0 --> '-', 1 --> 'Y', 2 --> 'X', ... , 25 --> 'B', 26 --> 'A' */
/*---------------------------------------------------------------------------*/
char decode(
  letter i /* letter to be decoded */
);

/*===========================================================================*/
/* -- READ_SCORES --
 * Read score matrix from 'fptr' into 'sc' */
/*---------------------------------------------------------------------------*/
void read_scores(
  FILE *fptr, /* input file pointer */
  scores *sc /* score matrix */
);

/*===========================================================================*/
/* -- WRITE_SCORES --
 * Write score matrix 'sc' as lower triangle matrix to 'fptr' */
/*---------------------------------------------------------------------------*/
void write_scores(
  FILE *fptr, /* output file pointer */
  scores *sc /* score matrix */
);

/*===========================================================================*/
/* -- WRITE_SCORES_MSA --
 * Write score matrix 'sc' in msa format to 'fptr' */
/*---------------------------------------------------------------------------*/
void write_scores_msa(
  FILE *fptr, /* output file pointer */
  scores *sc /* score matrix */
);

/*===========================================================================*/
/* -- RESET_GAPCOST --
 * Set the terminal gap score of 'sc' according to 'free_shift' */
/*---------------------------------------------------------------------------*/
void reset_gapcost(
  scores *sc, /* score matrix */
  BOOL free_shift /* free shift flag */
);

#endif /* SCORES_H */
/****** EOF (scores.h) *******************************************************/
